package io.xjar.boot;

import java.io.InputStream;
import java.security.CodeSource;
import java.security.CodeSigner;
import java.io.OutputStream;
import io.xjar.XKit;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.net.MalformedURLException;
import java.net.URLStreamHandler;
import io.xjar.reflection.XReflection;
import java.net.URLClassLoader;
import io.xjar.key.XKey;
import io.xjar.XEncryptor;
import io.xjar.XDecryptor;
import java.net.URL;
import java.lang.reflect.Method;
import org.springframework.boot.loader.LaunchedURLClassLoader;

public class XBootClassLoader extends LaunchedURLClassLoader
{
    private final XBootURLHandler xBootURLHandler;
    private final Object urlClassPath;
    private final Method getResource;
    private final Method getCodeSourceURL;
    private final Method getCodeSigners;
    
    public XBootClassLoader(final URL[] urls, final ClassLoader parent, final XDecryptor xDecryptor, final XEncryptor xEncryptor, final XKey xKey) throws Exception {
        super(urls, parent);
        this.xBootURLHandler = new XBootURLHandler(xDecryptor, xEncryptor, xKey, (ClassLoader)this);
        this.urlClassPath = XReflection.field((Class)URLClassLoader.class, "ucp").get((Object)this).value();
        this.getResource = XReflection.method((Class)this.urlClassPath.getClass(), "getResource", new Class[] { String.class }).method();
        this.getCodeSourceURL = XReflection.method((Class)this.getResource.getReturnType(), "getCodeSourceURL", new Class[0]).method();
        this.getCodeSigners = XReflection.method((Class)this.getResource.getReturnType(), "getCodeSigners", new Class[0]).method();
    }
    
    public URL findResource(final String name) {
        final URL url = super.findResource(name);
        if (url == null) {
            return null;
        }
        try {
            return new URL(url.getProtocol(), url.getHost(), url.getPort(), url.getFile(), (URLStreamHandler)this.xBootURLHandler);
        }
        catch (final MalformedURLException e) {
            return url;
        }
    }
    
    public Enumeration<URL> findResources(final String name) throws IOException {
        final Enumeration<URL> enumeration = super.findResources(name);
        if (enumeration == null) {
            return null;
        }
        return new XBootEnumeration(enumeration);
    }
    
    protected Class<?> findClass(final String name) throws ClassNotFoundException {
        try {
            return super.findClass(name);
        }
        catch (final ClassFormatError e) {
            final String path = name.replace('.', '/').concat(".class");
            final URL url = this.findResource(path);
            if (url == null) {
                throw new ClassNotFoundException(name, e);
            }
            try (InputStream in = url.openStream()) {
                final ByteArrayOutputStream bos = new ByteArrayOutputStream();
                XKit.transfer(in, (OutputStream)bos);
                final byte[] bytes = bos.toByteArray();
                final Object resource = this.getResource.invoke(this.urlClassPath, path);
                final URL codeSourceURL = (URL)this.getCodeSourceURL.invoke(resource, new Object[0]);
                final CodeSigner[] codeSigners = (CodeSigner[])this.getCodeSigners.invoke(resource, new Object[0]);
                final CodeSource codeSource = new CodeSource(codeSourceURL, codeSigners);
                return this.defineClass(name, bytes, 0, bytes.length, codeSource);
            }
            catch (final Throwable t) {
                throw new ClassNotFoundException(name, t);
            }
        }
    }
    
    static {
        ClassLoader.registerAsParallelCapable();
    }
    
    private class XBootEnumeration implements Enumeration<URL>
    {
        private final Enumeration<URL> enumeration;
        
        XBootEnumeration(final Enumeration<URL> enumeration) {
            this.enumeration = enumeration;
        }
        
        @Override
        public boolean hasMoreElements() {
            return this.enumeration.hasMoreElements();
        }
        
        @Override
        public URL nextElement() {
            final URL url = this.enumeration.nextElement();
            if (url == null) {
                return null;
            }
            try {
                return new URL(url.getProtocol(), url.getHost(), url.getPort(), url.getFile(), (URLStreamHandler)XBootClassLoader.this.xBootURLHandler);
            }
            catch (final MalformedURLException e) {
                return url;
            }
        }
    }
}
